/*
 * Timeserver
 *
 * References:
 *
 * Daytime Protocol
 * http://www.ietf.org/rfc/rfc0867.txt
 *
 * Time Protocol
 * http://www.ietf.org/rfc/rfc0868.txt
 *
 * NTP Protocol
 * http://www.ietf.org/rfc/rfc0958.txt
 * http://www.ietf.org/rfc/rfc1305.txt
 *
 * SNTP Protocol
 * http://www.ietf.org/rfc/rfc1361.txt
 * http://www.ietf.org/rfc/rfc1769.txt
 * http://www.ietf.org/rfc/rfc2030.txt
 * http://www.ietf.org/rfc/rfc4330.txt
 *
 */

#include "timesvr.h"

int32 julianDay(int16 cyr, int8 mo, int8 dy)
{
  int16 y;
  int8 m;
  y = cyr + 4800;
  if (mo < 3) {
    y--;
    m = mo + 9;
  } else {
    m = mo - 3;
  }
  return dy + (_mul(153, m) + 2) / 5 + (y / 4) - (y / 100) + (y / 400) + _mul(365, y) - 32045;
}

// calculate the seconds since midnight 1 Jan year to the given date/time (assumes same timezones)
// for year=1900:
// 2208988800 0x83AA7E80 corresponds to 00:00  1 Jan 1970 GMT
// 2398291200 0x8EF30500 corresponds to 00:00  1 Jan 1976 GMT
// 2524521600 0x96792480 corresponds to 00:00  1 Jan 1980 GMT
// 2629584000 0x9CBC4480 corresponds to 00:00  1 May 1983 GMT
int32 secondsSinceYear(int16 year, int16 cyr, int8 mo, int8 dy, int8 hr, int8 mn, int8 sc)
{
  return (julianDay(cyr, mo, dy) - julianDay(year, 1, 1)) * 86400 + _mul(hr, 3600) + _mul(mn, 60) + sc;
}

/* * * * * * * * * * * * * * *
 * NMEA RMC GPS Synchronizer *
 * * * * * * * * * * * * * * */

// Sample NMEA RMC messages:
/*
$GPRMC,123519,A,4807.038,N,01131.000,E,022.4,084.4,230394,003.1,W*6A
$GPRMC,183729,A,3907.356,N,12102.482,W,000.0,360.0,080301,015.5,E*6F
$GPRMC,183731,A,3907.482,N,12102.436,W,000.0,360.0,080301,015.5,E*67
$GPRMC,002454,A,3553.5295,N,13938.6570,E,0.0,43.1,180700,7.1,W,A*3F
$GPRMC,002456,A,3553.5295,N,13938.6570,E,0.0,43.1,180700,7.1,W,A*3D
$GPRMC,023042,A,3907.3837,N,12102.4684,W,0.0,156.1,131102,15.3,E,A*36
$GPRMC,023044,A,3907.3840,N,12102.4692,W,0.0,156.1,131102,15.3,E,A*37
$GPRMC,152926,V,6027.8259,N,02225.6713,E,10.8,0.0,190803,5.9,E,S*22
$GPRMC,152928,V,6027.8319,N,02225.6713,E,10.8,0.0,190803,5.9,E,S*29
$GPRMC,184050.84,A,3907.3839,N,12102.4772,W,00.0,000.0,080301,15,E*54
$GPRMC,184051.81,A,3907.3839,N,12102.4771,W,00.0,000.0,080301,15,E*53
$GPRMC,104715.20,A,5100.2111,N,00500.0006,E,21.7,003.0,140801,01.,W*70
$GPRMC,171537,A,3350.975,N,11823.991,W,0.0,096.5,060401,013.0,E,D*07
$GPRMC,230611.016,V,3907.3813,N,12102.4635,W,0.14,136.40,041002,,*04
$GPRMC,120557.916,A,5058.7456,N,00647.0515,E,0.00,82.33,220503,,*39
$GPRMC,001430.003,A,3907.3885,N,12102.4767,W,000.0,175.3,220403,015.4,E*71
$GPRMC,050306,V,4259.8839,N,07130.3922,W,010.3,139.7,291003,,*10
$GPRMC,162254.00,A,3723.02837,N,12159.39853,W,0.820,188.36,110706,,,A*74
*/

// time the last sync occurred with the GPS (for NTP reporting)
int32 lastsync_seconds = 0;
int32 lastsync_fracseconds = 0;

int8 gps_num[6]; // day, month, year, hour, minute, second
int16 gps_frac, gps_frac_divisor;
int1 gps_valid;
int32 last_message = 0;

// handle data gathered from RMC message
void timeserver_handle_rmc(void)
{
  int16 cyr, m;
  int8 yr, mo, dy, hr, mn, sc, start;
  status_led1 = gps_valid ? 100 : -100; // turn LED1 on for 100ms, green if valid position, red if not
  // make sure the fractional seconds are in milliseconds
  while (gps_frac_divisor < 1000)
  {
    gps_frac *= 10;
    gps_frac_divisor *= 10;
  }
  last_message = get_uptime();
  get_datetime(cyr, mo, dy, hr, mn, sc, m);
  start = get_millisup();
  yr = cyr % 100;
  // initial sanity check - years must be within one of each other
  if (((yr > gps_num[2]) ? (yr - gps_num[2]) : (gps_num[2] - yr)) <= 1)
  {
    int32 rtc_sec, gps_sec, diff;
    int16 gps_cyr;
    int1 rtc_fast;
    rtc_sec = secondsSinceYear(1900, cyr, mo, dy, hr, mn, sc);
    gps_cyr = cyr - (int16)yr + (int16)gps_num[2];
    // apply RMC delay correction
    if (!pps_adjust_millis(gps_frac) && TimeServerHasSync) return; // return if PPS sync failed and it isn't _too_ long since the last sync
    // if we got here, the time has been adjusted via the PPS, or the
    // sync age is more than the maximum allowed
    while (gps_frac >= 1000)
    {
      gps_frac -= 1000;
      if (++gps_num[5] == 60)
      {
        gps_num[5] = 0;
        if (++gps_num[4] == 60)
        {
          gps_num[4] = 0;
          if (++gps_num[3] == 24)
          {
            gps_num[3] = 0;
            if (++gps_num[0] > maxDom(gps_cyr, gps_num[1]))
            {
              gps_num[0] = 1;
              if (++gps_num[1] == 13)
              {
                gps_num[1] = 1;
                ++gps_num[2];
                ++gps_cyr;
              }
            }
          }
        }
      }
    }
    gps_sec = secondsSinceYear(1900, gps_cyr, gps_num[1], gps_num[0], gps_num[3], gps_num[4], gps_num[5]);
    // second sanity check - times must be close to each other
    rtc_fast = (rtc_sec > gps_sec || ((rtc_sec == gps_sec) && (m > gps_frac)));
    if (rtc_fast) diff = (rtc_sec - gps_sec) * 1000 + m - gps_frac;
    else          diff = (gps_sec - rtc_sec) * 1000 + gps_frac - m;
    if (diff == 0) return; // no need to make changes
    if (!gps_valid)
    {
      if (TimeServerHasSync) return;
      diff = 1; // allow small corrections if we've never sync'd
    }
    if (diff <= _mul(Config.TimeServerGPS_MaxDiff, 1000))
    {
      if (rtc_fast)
      {
        // need to subtract difference from RTC
        while (diff > 0)
        {
          if (diff <= m)
          {
            m -= diff;
            diff = 0;
          }
          else
          {
            if (diff >= 1000)
            {
              diff -= 1000;
            }
            else
            {
              m = 1000 + m - diff;
              diff = 0;
            }
            --rtc_sec;
            if (sc != 0) sc--;
            else
            {
              if (mn != 0) mn--;
              else
              {
                if (hr != 0) hr--;
                else
                {
                  if (dy != 1) dy--;
                  else
                  {
                    if (mo != 1) mo--;
                    else
                    {
                      mo = 12;
                      cyr--;
                    }
                    dy = maxDom(cyr, mo);
                  }
                  hr = 23;
                }
                mn = 59;
              }
              sc = 59;
            }
          }
        }
      }
      else
      {
        // need to add difference to RTC
        diff += m;
        while (diff >= 1000)
        {
          diff -= 1000;
          ++rtc_sec;
          if (++sc == 60)
          {
            sc = 0;
            if (++mn == 60)
            {
              mn = 0;
              if (++dy > maxDom(cyr, mo))
              {
                dy = 1;
                if (++mo == 13)
                {
                  mo = 1;
                  ++cyr;
                }
              }
            }
          }
        }
        m = diff;
      }
      // set the RTC to the GPS time
      m += ((int8)get_millisup() - start); // factor in processing time
      set_datetime(cyr, mo, dy, hr, mn, sc, m);
      if (gps_valid)
      {
        lastsync_seconds = rtc_sec;
        lastsync_fracseconds = (int32)m * 4294967L;
        TimeServerLastSync = get_uptime();
        TimeServerHasSync = TRUE;
      }
    }
  }
}

void timeserver_decode_datetime(int8 *nums, char *p)
{
  int8 i;
  for (i = 0; i < 3; i++) {
    *nums = (*p++ - '0') * 10;
    *nums++ += (*p++ - '0');
  }
}

// like strtok(buf,",") but doesn't skip consecutive instances of commas
// (so empty NMEA fields, eg ",,", aren't missed)
char *timeserver_nmea_tok_p;
char *timeserver_nmea_tok(char *buf)
{
  char *ret;
  if (buf) timeserver_nmea_tok_p = buf;
  ret = timeserver_nmea_tok_p;
  while (*timeserver_nmea_tok_p != ',' && *timeserver_nmea_tok_p != 0) timeserver_nmea_tok_p++;
  if (*timeserver_nmea_tok_p == ',') *timeserver_nmea_tok_p++ = 0;
  return ret;
}

void timeserver_handle_nmea_message(char *msg)
{
  char *p;
  int8 i;
  p = timeserver_nmea_tok(msg);
  if (p[0] != 'G' || p[1] != 'P') return;
  if (p[2] == 'R' && p[3] == 'M' && p[4] == 'C')
  {
    gps_frac = 0;
    gps_frac_divisor = 1;
    // time of day
    p = timeserver_nmea_tok(0);
    timeserver_decode_datetime(&gps_num[3], p);
    if (p[6] == '.')
    {
      for (i = 7; p[i] != 0 && gps_frac_divisor < 1000; ++i)
      {
        gps_frac = gps_frac * 10 + (p[i] - '0');
        gps_frac_divisor *= 10;
      }
    }
    // location validity flag
    p = timeserver_nmea_tok(0);
    gps_valid = (p[0] == 'A');
    // latitude
    p = timeserver_nmea_tok(0);
    strcpy(&TimeServerLatitude[1], p);
    TimeServerLatitude[0] = TimeServerLatitude[1];
    TimeServerLatitude[1] = TimeServerLatitude[2];
    TimeServerLatitude[2] = 0;
    // N/S hemisphere
    p = timeserver_nmea_tok(0);
    TimeServerLatHemi = p[0];
    // longitude
    p = timeserver_nmea_tok(0);
    strcpy(&TimeServerLongitude[1], p);
    TimeServerLongitude[0] = TimeServerLongitude[1];
    TimeServerLongitude[1] = TimeServerLongitude[2];
    TimeServerLongitude[2] = TimeServerLongitude[3];
    TimeServerLongitude[3] = 0;
    // E/W hemisphere
    p = timeserver_nmea_tok(0);
    TimeServerLongHemi = p[0];
    // speed
    timeserver_nmea_tok(0);
    // heading
    timeserver_nmea_tok(0);
    // date
    p = timeserver_nmea_tok(0);
    timeserver_decode_datetime(&gps_num[0], p);
    // not interested in any more tokens
    timeserver_handle_rmc();
  }
  else if (p[2] == 'G' && p[3] == 'S' && p[4] == 'V')
  {
    int8 msgnum;
    timeserver_nmea_tok(0); // total messages
    p = timeserver_nmea_tok(0); // message number (1..3)
    msgnum = *p - '0' - 1;
    p = timeserver_nmea_tok(0); // total SVs
    if (msgnum == 0)
      for (i = a_to_l(p); i < 12; i++)
        sat_id[i] = 0;
    for (i = 0; i < 4; i++)
    {
      p = timeserver_nmea_tok(0); // id
      sat_id[msgnum * 4 + i] = a_to_l(p);
      p = timeserver_nmea_tok(0); // elevation
      sat_elevation[msgnum * 4 + i] = a_to_l(p);
      p = timeserver_nmea_tok(0); // azimuth
      sat_azimuth[msgnum * 4 + i] = a_to_l(p);
      p = timeserver_nmea_tok(0); // snr
      sat_dbm[msgnum * 4 + i] = a_to_l(p);
    }
  }
}

char timeserver_serial_buf[80];
int8 timeserver_serial_cs;
void timeserver_handle_serial(void)
{
  static int8 cs_remaining = 0;
  static int8 bufidx = 0;
  while (bkbhit())
  {
    int8 c;
    c = bgetc();
    if (c == '$')
    {
      bufidx = 0;
      cs_remaining = 0;
    }
    else if (c == '*')
    {
      timeserver_serial_cs = 0;
      cs_remaining = 2;
    }
    else if (cs_remaining > 0)
    {
      timeserver_serial_cs = timeserver_serial_cs * 16 + hexchar(c);
      if (--cs_remaining == 0)
      {
        for (c = 0; c < bufidx; ++c) timeserver_serial_cs ^= timeserver_serial_buf[c];
        if (timeserver_serial_cs == 0)
        {
          timeserver_serial_buf[bufidx] = 0;
          timeserver_handle_nmea_message(timeserver_serial_buf);
        }
      }
    }
    else if (bufidx < (sizeof(timeserver_serial_buf) - 1))
    {
      timeserver_serial_buf[bufidx++] = c;
    }
  }
  {
    int32 ut;
    ut = get_uptime();
    if (ut - last_message > 5)
    {
      int8 i;
      // too long since last message - do a long LED pulse
      status_led1 = -2500; // turn LED1 red for 2500ms
      last_message = ut;
      for (i = 0; i < 12; i++) sat_id[i] = 0;
      // try a different baud rate
      i = serial_baud + 1;
      if (i > 7) i = 0;
      serial_setup_uart(i);
    }
  }
}

/* * * * * * * * * * *
 * TCP/IP Timeserver *
 * * * * * * * * * * */

#define NTP_VERSION 4

#if (DAYTIME_PORT != 0)
TCP_SOCKET daytime_tcp_socket;
UDP_SOCKET daytime_udp_socket;
#endif
#if (TIME_PORT != 0)
TCP_SOCKET time_tcp_socket;
UDP_SOCKET time_udp_socket;
#endif
#if (NTP_PORT != 0)
UDP_SOCKET ntp_socket;
#endif
#if (NBNS_PORT != 0)
UDP_SOCKET nbns_socket;
#endif
#if (NETBIOS_PORT != 0)
TCP_SOCKET netbios_socket;
#endif

void TimeServerInit(void)
{
  int8 i;
  TimeServerLastSync = 0;
  TimeServerHasSync = FALSE;
  TimeServerLatitude[0] = 0;
  TimeServerLatitude[3] = 0;
  TimeServerLatHemi = 'N';
  TimeServerLongitude[0] = 0;
  TimeServerLongitude[4] = 0;
  TimeServerLongHemi = 'E';
  for (i = 0; i < 12; ++i) sat_id[i] = 0;
  #if (DAYTIME_PORT != 0)
  daytime_tcp_socket = TCPListen(DAYTIME_PORT);
  daytime_udp_socket = UDPOpen(DAYTIME_PORT, NULL, INVALID_UDP_SOCKET);
  #endif
  #if (TIME_PORT != 0)
  time_tcp_socket = TCPListen(TIME_PORT);
  time_udp_socket = UDPOpen(TIME_PORT, NULL, INVALID_UDP_SOCKET);
  #endif
  #if (NTP_PORT != 0)
  ntp_socket = UDPOpen(NTP_PORT, NULL, INVALID_UDP_SOCKET);
  #endif
  #if (NBNS_PORT != 0)
  nbns_socket = UDPOpen(NBNS_PORT, NULL, INVALID_UDP_SOCKET);
  #endif
  #if (NETBIOS_PORT != 0)
  netbios_socket = TCPListen(NETBIOS_PORT);
  #endif
}

const char daytime_dayname[][*]={"Sunday","Monday","Tuesday","Wednesday","Thursday","Friday","Saturday"};
const char daytime_monname[][*]={"January","February","March","April","May","June","July","August","September","October","November","December"};

WORD timeserver_port;
int1 timeserver_ntp_broadcast;
int1 timeserver_udp;
BYTE timeserver_socket;

int8 timeserver_getc(void)
{
  int8 b;
  if (timeserver_udp) UDPGet(&b);
  else TCPGet(timeserver_socket, &b);
  return b;
}

void timeserver_skip(int8 n)
{
  while (n--) timeserver_getc();
}

void timeserver_putc(char c)
{
  if (timeserver_udp) UDPPut(c);
  else TCPPut(timeserver_socket, c);
}

int16 timeserver_getint16(void)
{
  int8 msb, lsb;
  msb = timeserver_getc();
  lsb = timeserver_getc();
  return make16(msb, lsb);
}

int16 timeserver_getint16le(void)
{
  int8 msb, lsb;
  lsb = timeserver_getc();
  msb = timeserver_getc();
  return make16(msb, lsb);
}

int32 timeserver_getint32(void)
{
  return make32(timeserver_getc(), timeserver_getc(), timeserver_getc(), timeserver_getc());
}

void timeserver_putint16(int16 v)
{
  timeserver_putc(make8(v, 1));
  timeserver_putc(make8(v, 0));
}

void timeserver_putint16le(int16 v)
{
  timeserver_putc(make8(v, 0));
  timeserver_putc(make8(v, 1));
}

void timeserver_putint32(int32 v)
{
  timeserver_putc(make8(v, 3));
  timeserver_putc(make8(v, 2));
  timeserver_putc(make8(v, 1));
  timeserver_putc(make8(v, 0));
}

void timeserver_putint32le(int32 v)
{
  timeserver_putc(make8(v, 0));
  timeserver_putc(make8(v, 1));
  timeserver_putc(make8(v, 2));
  timeserver_putc(make8(v, 3));
}

// This array is made of blocks of constant data. Each block starts with a byte
// count, followed by that many bytes.
const int8 timeserver_const_data[] = {
#define CONST_NTP_HEADER 0
#define CONST_NTP_HEADER_LEN 15
  CONST_NTP_HEADER_LEN,
      1,           // stratum
     10,           // 2^10 = 1024sec polling interval
    -10,           // 2^-10 = ~1ms precision
    0,1,0,0,       // root delay (1s)
    0,0,128,0,     // root dispersion (~0.5s)
    'G','P','S',0, // reference identifier
#define CONST_NTP_BCAST_TRAILER (CONST_NTP_HEADER+CONST_NTP_HEADER_LEN+1)
#define CONST_NTP_BCAST_TRAILER_LEN 8
  CONST_NTP_BCAST_TRAILER_LEN,
    0,0,0,0,0,0,0,0, // receive timestamp
#define CONST_NBNS_HDR1 (CONST_NTP_BCAST_TRAILER+CONST_NTP_BCAST_TRAILER_LEN+1)
#define CONST_NBNS_HDR1_LEN 4
  CONST_NBNS_HDR1_LEN,
    0x01,0x10, // flags: must be broadcast name query
    0,1,       // qs: can only handle one query at a time
#define CONST_NBNS_HDR2 (CONST_NBNS_HDR1+CONST_NBNS_HDR1_LEN+1)
#define CONST_NBNS_HDR2_LEN 5
  CONST_NBNS_HDR2_LEN,
    0,      // trailing zero
    0,0x20, // type: must be NB
    0,1,    // class: must be IN
#define CONST_NBNS_RSP1 (CONST_NBNS_HDR2+CONST_NBNS_HDR2_LEN+1)
#define CONST_NBNS_RSP1_LEN 11
  CONST_NBNS_RSP1_LEN,
    0x85,0,
    0,0, // questions
    0,1, // answer rrs
    0,0, // auth rrs
    0,0, // addr rrs
    0x20, // name
#define CONST_NBNS_RSP2 (CONST_NBNS_RSP1+CONST_NBNS_RSP1_LEN+1)
#define CONST_NBNS_RSP2_LEN 13
  CONST_NBNS_RSP2_LEN,
    0,
    0,0x20, // type = NB
    0,0x01, // class = IN
    0,0,    // TTL
    0,0,    // TTL
    0,6,    // rdlength
    0,0,    // not a group, B node
#define CONST_SMB_HDR1 (CONST_NBNS_RSP2+CONST_NBNS_RSP2_LEN+1)
#define CONST_SMB_HDR1_LEN 3
  CONST_SMB_HDR1_LEN,
    0, // message type: session message
    0, // flags
    0, // length MSB (16-bit, big endian)
#define CONST_SMB_HDR2 (CONST_SMB_HDR1+CONST_SMB_HDR1_LEN+1)
#define CONST_SMB_HDR2_LEN 4
  CONST_SMB_HDR2_LEN,
    0xFF,'S','M','B', // SMB header
#define CONST_SMB_HDR3 (CONST_SMB_HDR2+CONST_SMB_HDR2_LEN+1)
#define CONST_SMB_HDR3_LEN 16
  CONST_SMB_HDR3_LEN,
    0,    // error code MSB (16-bit, little endian)
    0x98, // flags
    0,0,  // flags2
    0,0,0,0,0,0,0,0,0,0,0,0, // extra
#define CONST_NB_SR_RESP (CONST_SMB_HDR3+CONST_SMB_HDR3_LEN+1)
#define CONST_NB_SR_RESP_LEN 4
  CONST_NB_SR_RESP_LEN,
    0x82, // positive session response
    0,    // flags
    0,0,  // length
#define CONST_NEG_RESP1 (CONST_NB_SR_RESP+CONST_NB_SR_RESP_LEN+1)
#define CONST_NEG_RESP1_LEN 14
  CONST_NEG_RESP1_LEN,
    0,0, // security mode: share
    0,4, // max transmit buffer size (0x0400=1024)
    1,0, // max multiplexed requests (1)
    1,0, // max vcs between client and server (1)
    0,0, // raw modes supported (none)
    0,0,0,0, // session key
#define CONST_NEG_RESP2 (CONST_NEG_RESP1+CONST_NEG_RESP1_LEN+1)
#define CONST_NEG_RESP2_LEN 4
  CONST_NEG_RESP2_LEN,
    0,0, // encryption key length
    0,0, // reserved
#define CONST_SESSION_SETUP (CONST_NEG_RESP2+CONST_NEG_RESP2_LEN+1)
#define CONST_SESSION_SETUP_LEN 26
  CONST_SESSION_SETUP_LEN,
    3,     // word count
    0x75,  // andx command tree connect
    0,     // reserved
    44,0,  // andx offset
    0,0,   // action
    3,0,   // byte count
    0,0,0, // **** the workgroup name is one of these
    3,     // word count
    0xFF,  // andx command (no further commands)
    0,     // reserved
    0,0,   // offset
    0,0,   // optional support
    5,0,   // byte count
    'I','P','C',0, // IPC
    0,     // native file system
#define CONST_TREECON_RESP (CONST_SESSION_SETUP+CONST_SESSION_SETUP_LEN+1)
#define CONST_TREECON_RESP_LEN 7
  CONST_TREECON_RESP_LEN,
    2,   // word count
    0,4, // max buffer size (0x0400=1024)
    0,0, // tree id
    0,0, // byte count
#define CONST_OPEN_RESP (CONST_TREECON_RESP+CONST_TREECON_RESP_LEN+1)
#define CONST_OPEN_RESP_LEN 3
#define CONST_TREEDIS_RESP CONST_OPEN_RESP
  CONST_OPEN_RESP_LEN,
    0,   // word count
    0,0, // byte count
#define CONST_TOD_RESP (CONST_OPEN_RESP+CONST_OPEN_RESP_LEN+1)
#define CONST_TOD_RESP_LEN 28
  CONST_TOD_RESP_LEN,
    10,   // word count
    4,0,  // total parameter count
    21,0, // total data count
    0,0,  // reserved
    4,0,  // parameter count
    56,0, // parameter offset
    0,0,  // parameter displacement
    21,0, // data count
    60,0, // data offset
    0,0,  // data displacement
    0,    // byte count
    0,    // pad
    26,0, // byte count
    0,    // pad
    0,0,  // error code
    0,0,  // convertor
#define CONST_TOD_REQ (CONST_TOD_RESP+CONST_TOD_RESP_LEN+1)
#define CONST_TOD_REQ_LEN 18
  CONST_TOD_REQ_LEN,
    0x5B,0, // rap opcode
    'r','L',0, // parameter descriptor
    'D','D','B','B','B','B','W','W','B','B','W','B',0, // return descriptor
#define CONST_END (CONST_TOD_REQ+CONST_TOD_REQ_LEN+1)
};
#if (CONST_END > 255)
#error constant index must be int16
#endif

// put constant data
void timeserver_put_const(int8 idx)
{
  int8 count;
  count = timeserver_const_data[idx++];
  while (count--) timeserver_putc(timeserver_const_data[idx++]);
}

// get constant data
// returns TRUE if all the received data matched the constant
int1 timeserver_get_const(int8 idx)
{
  int8 count;
  int1 ret = TRUE;
  count = timeserver_const_data[idx++];
  while (count--) if (timeserver_getc() != timeserver_const_data[idx++]) ret = FALSE;
  return ret;
}

void timeserver_handle_request(void)
{
  int8 ntp_originate_flags;
  int32 ntp_originate_seconds;
  int32 ntp_originate_fracseconds;
  if (timeserver_ntp_broadcast || (timeserver_udp ? UDPIsGetReady(timeserver_socket) : TCPIsConnected(timeserver_socket)))
  {
    // got a request of some sort
    if (timeserver_port == NTP_PORT)
    {
      if (timeserver_ntp_broadcast)
      {
        ntp_originate_flags = 5; // broadcast
      }
      else
      {
        // need originate flags & timestamp
        int8 i;
        ntp_originate_flags = timeserver_getc();
        for (i = 0; i < 39; i++) timeserver_getc();;
        // get seconds part
        ntp_originate_seconds = timeserver_getint32();
        // get fractional part
        ntp_originate_fracseconds = timeserver_getint32();
      }
    }
    if (timeserver_udp)
    {
      UDPDiscard();
    }
    else
    {
      TCPDiscard(timeserver_socket);
    }
    if (timeserver_udp ? UDPIsPutReady(timeserver_socket) : TCPIsPutReady(timeserver_socket))
    {
      int32 seconds;
      int32 fracseconds; // fractional seconds (2**-32 second units)
      int16 cyr, m;
      int8 mo, dy, dw, hr, mn, sc;
      get_datetime(cyr, mo, dy, hr, mn, sc, m);
      dw = dayOfWeek(cyr, mo, dy);
      seconds = secondsSinceYear(1900, cyr, mo, dy, hr, mn, sc);
      switch (timeserver_port)
      {
        case DAYTIME_PORT:
          printf(timeserver_putc, "%s, %s %02u, %02u%02u %02u:%02u:%02u-GMT",
                 daytime_dayname[dw], daytime_monname[mo - 1], dy,
                 (int8)(cyr / 100), (int8)(cyr % 100), hr, mn, sc);
          break;
        case TIME_PORT:
          timeserver_putint32(seconds);
          break;
        case NTP_PORT:
          fracseconds = (int32)m * 4294967L; // so that millis (0..1000) scales to 0..2^32
               if ((ntp_originate_flags & 7) == 1) // 1=symmetric active
            ntp_originate_flags = 2;               // 2=symmetric passive
          else if ((ntp_originate_flags & 7) == 3) // 3=client
            ntp_originate_flags = 4;               // 4=server
          else if (!timeserver_ntp_broadcast)
            break; // unrecognized mode - ignore request
          ntp_originate_flags |= (NTP_VERSION<<3);
          if (!TimeServerHasSync) ntp_originate_flags |= 0xC0; // alarm state - clock unsynchronized
          timeserver_putc(ntp_originate_flags);
          timeserver_put_const(CONST_NTP_HEADER);
          timeserver_putint32(lastsync_seconds); // reference timestamp (time of last correction)
          timeserver_putint32(lastsync_fracseconds);
          if (timeserver_ntp_broadcast)
          {
            timeserver_putint32(seconds); // originate timestamp
            timeserver_putint32(fracseconds);
            timeserver_put_const(CONST_NTP_BCAST_TRAILER); // receive timestamp
          }
          else
          {
            timeserver_putint32(ntp_originate_seconds); // originate timestamp
            timeserver_putint32(ntp_originate_fracseconds);
            timeserver_putint32(seconds); // receive timestamp
            timeserver_putint32(fracseconds);
          }
          // transmit timestamp
          timeserver_putint32(seconds); // transmit timestamp
          timeserver_putint32(fracseconds);
          break;
      }
      status_led2 = 100; // turn LED2 off for 100ms
      if (timeserver_udp)
      {
        UDPFlush();
      }
      else
      {
        TCPFlush(timeserver_socket);
        TCPDisconnect(timeserver_socket);
      }
    }
  }
}

#if (NBNS_PORT != 0)

void nbns_handle_request(void)
{
  int8 i, i1, i2, b[17];
  int16 tid;
  int1 reqgood = TRUE;
  if (!UDPIsGetReady(timeserver_socket)) return;
  tid = timeserver_getint16(); // transaction ID
  if (!timeserver_get_const(CONST_NBNS_HDR1)) reqgood = FALSE;
  timeserver_skip(6); // answer rrs, auth rrs, add rrs
  if (timeserver_getc() != 0x20) reqgood = FALSE; // nq: must be a name query request
  for (i = 0; i < 16; i++)
  {
    int8 msb, lsb;
    msb = timeserver_getc();
    lsb = timeserver_getc();
    b[i] = (msb - 'A') * 16 + (lsb - 'A');
  }
  b[i] = 0;
  if (!timeserver_get_const(CONST_NBNS_HDR2)) reqgood = FALSE;
  UDPDiscard();
  if (!reqgood) return;
  
  // compare Config.NetBIOSName
  // case-insensitive, 16 character, space padded
  i1 = 0;
  i2 = 0;
  for (i = 0; i < 16; i++)
  {
    int8 c1, c2;
    c1 = Config.NetBIOSName[i1];
    c2 = b[i2];
    if (c1 == 0) c1 = ' ';
    else
    {
      i1++;
      if (c1 >= 'a' && c1 <= 'z') c1 -= ('a' - 'A');
    }
    if (c2 == 0) c2 = ' ';
    else
    {
      i2++;
      if (c2 >= 'a' && c2 <= 'z') c2 -= ('a' - 'A');
    }
    if (c1 != c2) break;
  }
  if (i == 16)
  {
    // name matches
    timeserver_putint16(tid);
    timeserver_put_const(CONST_NBNS_RSP1);
    for (i = 0; i < 16; i++)
    {
      timeserver_putc((b[i] >> 4) + 'A');
      timeserver_putc((b[i] & 15) + 'A');
    }
    timeserver_put_const(CONST_NBNS_RSP2);
    timeserver_putint32le(Config.MyIPAddr);
    UDPFlush();
  }
}

#endif /* NBNS_PORT */

#if (NETBIOS_PORT != 0)

// Globals used to pass header variables
int8 netbios_hdr_length;
int8 netbios_hdr_cmd;
int8 netbios_hdr_errorClass;
int8 netbios_hdr_errorCode;
int8 netbios_hdr_ids[8];

// Send generic NetBIOS SMB header
void netbios_send_header(void)
{
  int8 i;
  timeserver_put_const(CONST_SMB_HDR1); // message type, flags, length MSB
  timeserver_putc(netbios_hdr_length);
  timeserver_put_const(CONST_SMB_HDR2); // SMB header
  timeserver_putc(netbios_hdr_cmd); // negotiate protocol
  timeserver_putc(netbios_hdr_errorClass); // error: success
  timeserver_putc(0); // reserved
  timeserver_putc(netbios_hdr_errorCode); // LSB error code (16-bit, little endian)
  timeserver_put_const(CONST_SMB_HDR3); // MSB error code, flags, flags2, extra
  for (i = 0; i < 8; i++) timeserver_putc(netbios_hdr_ids[i]); // ids
}

#define NUM_DIALECTS 6
const int8 dialects[NUM_DIALECTS][*] = {
  "PC NETWORK PROGRAM 1.0",
  "MICROSOFT NETWORKS 1.03",
  "MICROSOFT NETWORKS 3.0",
  "LANMAN1.0",
  "DOS LM1.2X002",
  "LM1.2X002"
};

// Assumes the timerserver TCP input stream is about to read the byte count for
// the block of dialect data.
// Returns the index to the most recent client dialect that we support, or -1
// if no dialect was found.
signed int8 netbios_select_dialect(void)
{
  int8 dialects_idx[NUM_DIALECTS];
  int16 i, count;
  signed int8 cur_dialect, sel_dialect;
  int8 idx;
  int1 state;
  count = timeserver_getint16le(); // byte count
  // parse through sent dialects, looking for the best match
  // states: expecting buffer format (0x02)
  //         stepping through null-terminated string
  state = 0;
  cur_dialect = -1;
  sel_dialect = -1;
  for (i = 0; i < count; i++)
  {
    int8 c;
    c = timeserver_getc();
    if (state == 0)
    {
      state = 1;
      if (c != 2)
      {
        sel_dialect = -1;
        timeserver_skip(count - i - 1);
        break;
      }
      // reset dialect string matchers
      cur_dialect++;
      for (idx = 0; idx < NUM_DIALECTS; idx++)
      {
        dialects_idx[idx] = 0;
      }
    }
    else
    {
      if (c == 0) state = 0;
      for (idx = 0; idx < NUM_DIALECTS; idx++)
      {
        int8 t;
        t = dialects_idx[idx];
        if (t != -1)
        {
          if (c == dialects[idx][t])
          {
            t++;
            if (c == 0)
            {
              t = -1;
              sel_dialect = cur_dialect;
            }
          }
          else
          {
            // dialect string doesn't match
            t = -1;
          }
        }
        dialects_idx[idx] = t;
      }
    }
  }
  return sel_dialect;
}

void netbios_handle_request(void)
{
  static int32 last_get = 0;
  int1 reqgood = TRUE;
  int32 now;
  if (!TCPIsConnected(timeserver_socket)) return;
  now = get_millisup();
  if (!TCPIsGetReady(timeserver_socket))
  {
    if (last_get != 0 && ((now - last_get) > 100))
    {
      TCPDisconnect(timeserver_socket);
      last_get = 0;
    }
    return;
  }
  last_get = now;
  netbios_hdr_errorClass = 0;
  netbios_hdr_errorCode = 0;
  switch (timeserver_getc()) // type
  {
    case 0x81:
      // session request
      TCPDiscard(timeserver_socket);
      timeserver_put_const(CONST_NB_SR_RESP); // positive session response, flags, length
      TCPFlush(timeserver_socket);
      break;
    case 0x00:
      // session message
      timeserver_skip(3); // flags, length
      if (timeserver_get_const(CONST_SMB_HDR2))
      {
        int8 count;
        netbios_hdr_cmd = timeserver_getc();
        timeserver_skip(19); // status, flags, flags2, extra
        for (count = 0; count < 8; count++) netbios_hdr_ids[count] = timeserver_getc(); // ids
        switch (netbios_hdr_cmd)
        {
          case 0x72:
            // negotiate request
            {
              signed int8 sel_dialect;
              int1 iscoreprotocol;
              count = timeserver_getc(); // word count
              timeserver_skip(count * 2);
              sel_dialect = netbios_select_dialect();
              TCPDiscard(timeserver_socket);
              iscoreprotocol = (sel_dialect == 0 || sel_dialect == -1);
              // send reply
              netbios_hdr_length = iscoreprotocol ? 37 : 61;
              netbios_send_header();
              if (iscoreprotocol)
              {
                timeserver_putc(1); // word count
                timeserver_putint16le((signed int16)sel_dialect); // selected dialect
              }
              else
              {
                timeserver_putc(13); // word count
                timeserver_putint16le((signed int16)sel_dialect); // selected dialect
                timeserver_put_const(CONST_NEG_RESP1);
                {
                  int16 m;
                  signed int16 cyr;
                  signed int8 mo, dy, hr, mn, sc, tz;
                  get_datetime(cyr, mo, dy, hr, mn, sc, m);
                  tz = RtcGetTimezone();
                  applyTimezone(cyr, mo, dy, hr, mn, sc, tz);
                  // server time - hhhhhmmm mmmsssss
                  timeserver_putc((sc >> 1) | (mn << 5));
                  timeserver_putc((hr << 3) | (mn >> 3));
                  // server date - yyyyyyym mmmddddd
                  timeserver_putc((mo << 5) | dy);
                  timeserver_putc(((int8)(cyr - 1980) << 1) | (mo >> 3));
                  // server timezone
                  timeserver_putint16le((signed int16)tz * -15);
                }
                timeserver_put_const(CONST_NEG_RESP2);
              }
              timeserver_putint16le(0); // byte count
              TCPFlush(timeserver_socket);
            }
            break;
          case 0x73:
            // session setup andx
            timeserver_getc(); // word count
            if (timeserver_getc() != 0x75) reqgood = FALSE; // andx command not a tree connect
            TCPDiscard(timeserver_socket);
            if (!reqgood) return;
            // send reply
            netbios_hdr_length = 58;
            netbios_send_header();
            timeserver_put_const(CONST_SESSION_SETUP);
            TCPFlush(timeserver_socket);
            break;
          case 0x70:
            // tree connect request
            TCPDiscard(timeserver_socket);
            netbios_hdr_length = 39;
            netbios_send_header();
            timeserver_put_const(CONST_TREECON_RESP);
            TCPFlush(timeserver_socket);
            break;
          case 0x2d:
            // open andx request
            TCPDiscard(timeserver_socket);
            netbios_hdr_length = 35;
            netbios_hdr_errorClass = 1; // dos
            netbios_hdr_errorCode = 2; // file not found
            netbios_send_header();
            timeserver_put_const(CONST_OPEN_RESP);
            TCPFlush(timeserver_socket);
            break;
          case 0x25:
            // trans
            // TODO make sure we're replying to the correct command
            count = timeserver_getc(); // word count
            timeserver_skip(count * 2);
            count = timeserver_getint16le(); // byte count
            timeserver_skip(count - 20);
            if (!timeserver_get_const(CONST_TOD_REQ)) reqgood = FALSE;
            if (timeserver_getint16le() < 21) reqgood = FALSE; // response buffer length
            TCPDiscard(timeserver_socket);
            if (!reqgood) return;
            // send reply
            netbios_hdr_length = 81;
            netbios_send_header();
            timeserver_put_const(CONST_TOD_RESP);
            {
              int16 m;
              signed int16 cyr;
              signed int8 mo, dy, hr, mn, sc, tz;
              get_datetime(cyr, mo, dy, hr, mn, sc, m);
              tz = RtcGetTimezone();
              applyTimezone(cyr, mo, dy, hr, mn, sc, tz);
              // time since 1970
              timeserver_putint32le(secondsSinceYear(1970, cyr, mo, dy, hr, mn, sc));
              // milliseconds up
              timeserver_putint32le(now);
              // time
              timeserver_putc(hr);
              timeserver_putc(mn);
              timeserver_putc(sc);
              timeserver_putc((int8)(m / 10));
              // timezone
              timeserver_putint16le((signed int16)tz * -15);
              // clock resolution (0.00001/sec)
              timeserver_putint16le(10); // 10=1ms
              // date
              timeserver_putc(dy);
              timeserver_putc(mo);
              timeserver_putint16le(cyr);
              // day of week
              timeserver_putc(dayOfWeek(cyr, mo, dy));
            }
            status_led2 = 100; // turn LED2 off for 100ms
            TCPFlush(timeserver_socket);
            break;
          case 0x71:
            // tree disconnect request
            TCPDiscard(timeserver_socket);
            netbios_hdr_length = 35;
            netbios_send_header();
            timeserver_put_const(CONST_TREEDIS_RESP);
            TCPFlush(timeserver_socket);
            break;
          default:
            TCPDiscard(timeserver_socket);
            break;
        }
      }
      else
      {
        TCPDiscard(timeserver_socket);
      }
      break;
    default:
      TCPDiscard(timeserver_socket);
      break;
  }
}

#endif /* NETBIOS_PORT */

#define DO_TIMESERVER_REQUEST(UDP, SOCKET) { timeserver_udp = UDP; timeserver_socket = SOCKET; timeserver_handle_request(); }
void TimeServerTask(void)
{
  static int32 TimeServerLastBroadcast;
  timeserver_handle_serial();
  timeserver_ntp_broadcast = 0;
  #if (DAYTIME_PORT != 0)
  timeserver_port = DAYTIME_PORT;
  if (daytime_tcp_socket != INVALID_SOCKET    ) DO_TIMESERVER_REQUEST(FALSE, daytime_tcp_socket);
  if (daytime_udp_socket != INVALID_UDP_SOCKET) DO_TIMESERVER_REQUEST(TRUE , daytime_udp_socket);
  #endif
  #if (TIME_PORT != 0)
  timeserver_port = TIME_PORT;
  if (time_tcp_socket != INVALID_SOCKET    ) DO_TIMESERVER_REQUEST(FALSE, time_tcp_socket);
  if (time_udp_socket != INVALID_UDP_SOCKET) DO_TIMESERVER_REQUEST(TRUE , time_udp_socket);
  #endif
  #if (NTP_PORT != 0)
  timeserver_port = NTP_PORT;
  if (ntp_socket != INVALID_UDP_SOCKET) DO_TIMESERVER_REQUEST(1, ntp_socket);
  if (Config.TimeServer_BroadcastInterval != 0)
  {
    int32 now;
    now = get_uptime();
    if (now - TimeServerLastBroadcast > (int32)Config.TimeServer_BroadcastInterval)
    {
      UDP_SOCKET s;
      NODE_INFO remote;
      remote.MACAddr.v[0] =
      remote.MACAddr.v[1] =
      remote.MACAddr.v[2] =
      remote.MACAddr.v[3] =
      remote.MACAddr.v[4] =
      remote.MACAddr.v[5] = 0xFF;
      remote.IPAddr.Val = AppConfig.MyIPAddr.Val | ~AppConfig.MyMask.Val;
      s = UDPOpen(NTP_PORT, &remote, NTP_PORT);
      if (s != INVALID_UDP_SOCKET)
      {
        timeserver_ntp_broadcast = TRUE;
        DO_TIMESERVER_REQUEST(TRUE, s);
        UDPClose(s);
        TimeServerLastBroadcast = now;
      }
    }
  }
  #endif
  #if (NBNS_PORT != 0)
  timeserver_port = NBNS_PORT;
  if (nbns_socket != INVALID_UDP_SOCKET)
  {
    timeserver_udp = TRUE;
    timeserver_socket = nbns_socket;
    nbns_handle_request();
  }
  #endif
  #if (NETBIOS_PORT != 0)
  timeserver_port = NETBIOS_PORT;
  if (netbios_socket != INVALID_SOCKET)
  {
    timeserver_udp = FALSE;
    timeserver_socket = netbios_socket;
    netbios_handle_request();
  }
  #endif
}
